# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from PyQt4 import QtCore, QtGui

from bzrlib import osutils

from bzrlib.plugins.explorer.lib import (
    app_suite,
    explorer_profile,
    kinds,
    )
from bzrlib.plugins.explorer.lib.i18n import gettext, N_
from bzrlib.plugins.explorer.widgets import tab_widget


# Preferences supported are:
# * app-suite - the application suite to use
# * toolbar-style - text-only, icon-only, text-beside-icon, text-under-icon
#
DEFAULT_PREFERENCES = {
    "app-suite": app_suite.app_suite_registry.default_key,
    "toolbar-contents": "default",
    "toolbar-style": "text-under-icon",
    "status-auto-refresh": True,
    "custom-dialog-init": True,
    "custom-dialog-branch": True,
    "custom-dialog-checkout": True,
    "advanced-commands-terminal": False,
    "workingtree-style": "qbrowse",
    }
_BOOLEAN_PREFERENCES = [
    "status-auto-refresh",
    "custom-dialog-init",
    "custom-dialog-branch",
    "custom-dialog-checkout",
    "advanced-commands-terminal",
    ]


def load_preferences():
    """Load and return the user preferences.

    :return: a dictionary of preferences
    """
    result = DEFAULT_PREFERENCES.copy()
    settings = _get_settings()
    settings.beginGroup("Preferences")
    for key in settings.allKeys():
        name = str(key)
        if name in _BOOLEAN_PREFERENCES:
            value = settings.value(name).toBool()
        else:
            value = str(settings.value(name).toString())
        result[name] = value
    settings.endGroup()
    # XXX: Should we trap errors above and return default settings if one occurs?
    return result


def save_preferences(preferences):
    """Save preferences.

    :param preferences: a dictionary of preferences to save
    """
    settings = _get_settings()
    settings.beginGroup("Preferences")
    for name in sorted(preferences):
        value = QtCore.QVariant(preferences[name])
        settings.setValue(name, value)
    settings.endGroup()


_settings = None


def _get_settings():
    """Return the QSettings object to use for preference loading/saving."""
    global _settings
    if _settings is None:
        config_dir = explorer_profile.config_dir()
        config_file = osutils.pathjoin(config_dir, "explorer.conf")
        _settings = QtCore.QSettings(config_file, QtCore.QSettings.IniFormat)
    return _settings


class ExplorerPreferencesDialog(QtGui.QDialog):

    def __init__(self, preferences, callback, title=None, parent=None):
        """Create the dialog.

        :param preferences: a dictionary of preference names and values
        :param callback: a callable to call when a preference is changed.
          The required signature is callback(name, new_value, old_value).
        :param title: the dialog title. None for "Preferences" (localised).
        :param parent: the parent Window. If not None, the dialog will be
          centered over this, otherwise the positioning is undefined.
        """
        QtGui.QDialog.__init__(self, parent)
        self._preferences = preferences
        self._callback = callback
        self._build_ui(title)
        # NOTE: It would be nice to restore the previous position and size.
        # QBzrDialog/QBzrWindow has support for the latter but doesn't
        # seem to restore the position (which is the interesting bit).
        # It's probably more appropriate to store them in our own settings
        # file anyhow, now that it exists

    def _build_ui(self, title=None):
        if title is None:
            title = gettext("Preferences")
        self.setWindowTitle(title)
        container = tab_widget.QBzrPreferencesTabWidget()
        self._add_section(container, gettext("Appearance"),
            kinds.APPEARANCE_PREFS, self._build_appearance)
        self._add_section(container, gettext("Behavior"),
            kinds.BEHAVIOUR_PREFS, self._build_behaviour)
        self._add_section(container, gettext("Applications"),
            kinds.APPLICATION_PREFS, self._build_applications)
        layout = QtGui.QVBoxLayout()
        layout.addWidget(container)
        self.setLayout(layout)

    def _add_section(self, container, title, icon_kind, layout_callable):
        widget= QtGui.QWidget()
        widget.setLayout(layout_callable())
        icon = kinds.icon_for_kind(icon_kind)
        container.addTab(widget, icon, title)

    def _build_applications(self):
        items = app_suite.app_suite_registry.keys()
        suite_combo = self._build_combo_field("app-suite", items)
        custom_dialog_init_cb = self._build_checkbox_field("custom-dialog-init",
            gettext("Use custom initialize dialog"))
        custom_dialog_branch_cb = self._build_checkbox_field("custom-dialog-branch",
            gettext("Use custom branch dialog"))
        custom_dialog_checkout_cb = self._build_checkbox_field("custom-dialog-checkout",
            gettext("Use custom checkout dialog"))
        layout = QtGui.QFormLayout()
        layout.addRow(gettext("Suite"), suite_combo)
        # Make the checkboxes span both columns. If this violates OS X
        # guidelines, we should change the behaviour on the platform.
        layout.addRow(custom_dialog_init_cb)
        layout.addRow(custom_dialog_branch_cb)
        layout.addRow(custom_dialog_checkout_cb)
        return layout

    def _build_appearance(self):
        # Toolbar
        tb_contents_combo = self._build_combo_field("toolbar-contents", [
            "default",
            "expanded",
            ])
        tb_style_combo = self._build_combo_field("toolbar-style", [
            "text-under-icon",
            "text-beside-icon",
            "icon-only",
            "text-only",
            ])
        tb_layout = QtGui.QFormLayout()
        tb_layout.addRow(gettext("Contents"), tb_contents_combo)
        tb_layout.addRow(gettext("Style"), tb_style_combo)
        tb_group = QtGui.QGroupBox(gettext("Toolbar"))
        tb_group.setLayout(tb_layout)

        # Working Tree
        wt_style_combo = self._build_combo_field("workingtree-style", [
            "qbrowse",
            "classic",
            ])
        wt_layout = QtGui.QFormLayout()
        wt_layout.addRow(gettext("Style"), wt_style_combo)
        wt_group = QtGui.QGroupBox(gettext("Working Tree"))
        wt_group.setLayout(wt_layout)

        layout = QtGui.QVBoxLayout()
        layout.addWidget(tb_group)
        layout.addWidget(wt_group)
        return layout

    def _build_behaviour(self):
        auto_refresh_cb = self._build_checkbox_field("status-auto-refresh",
            gettext("Automatically refresh status report"))
        terminal_cb = self._build_checkbox_field("advanced-commands-terminal",
            gettext("Open terminal shell for advanced commands"))
        layout = QtGui.QVBoxLayout()
        layout.addWidget(auto_refresh_cb)
        layout.addWidget(terminal_cb)
        layout.addStretch()
        return layout

    def _build_checkbox_field(self, pref_name, text):
        default_value = DEFAULT_PREFERENCES.get(pref_name)
        current_value = self._preferences.get(pref_name, default_value)
        checkbox = QtGui.QCheckBox(text)
        checkbox.setChecked(current_value)

        # Set-up change detection
        def checkbox_changed(checked):
            value = checked == QtCore.Qt.Checked
            old_value = self._preferences.get(pref_name)
            self._preferences[pref_name] = value
            self._callback(pref_name, value, old_value)
            save_preferences(self._preferences)
        self.connect(checkbox, QtCore.SIGNAL("stateChanged(int)"),
            checkbox_changed)
        return checkbox

    def _build_combo_field(self, pref_name, pref_values):
        combo = QtGui.QComboBox()
        # Set-up options and current selection
        combo.insertItems(0, pref_values)
        default_value = DEFAULT_PREFERENCES.get(pref_name)
        current_value = self._preferences.get(pref_name, default_value)
        try:
            current_index = pref_values.index(current_value)
        except ValueError:
            # The current value is no longer legal. That can happen when
            # the registry underlying the combo model has items removed
            # thanks to plugins being uninstalled, say. Fall back to the
            # default.
            current_index = pref_values.index(default_value)
            self._preferences[pref_name] = default_value
        combo.setCurrentIndex(current_index)

        # Set-up change detection
        def combo_changed(value):
            # value is a QVariant that gets reused so we need a copy here
            value = str(value)
            old_value = self._preferences.get(pref_name)
            self._preferences[pref_name] = value
            self._callback(pref_name, value, old_value)
            save_preferences(self._preferences)
        self.connect(combo, QtCore.SIGNAL("currentIndexChanged(QString)"),
            combo_changed)
        return combo
