# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import os
from bzrlib import config, osutils
from bzrlib.plugins.explorer.lib import bookmarks, tools
from bzrlib.plugins.explorer.lib.i18n import gettext, N_


# Names of data files within a profile directory
BOOKMARKS_FILENAME = "bookmarks.xml"
TOOLS_FILENAME = "tools.xml"
EDITORS_FILENAME = "editors.conf"

_DEFAULT_CONTENT = {
        BOOKMARKS_FILENAME: """
<folder title="My Bookmarks">
</folder>
""",
        TOOLS_FILENAME: """
<folder title="My Tools">
</folder>
""",
    }


class ExplorerProfile(object):

    def __init__(self, directory, title):
        """Construct a profile for Bazaar Explorer.

        :param directory: the directory to load the profile from or
          create the profile in.
        :param title: title of profile
        """
        #print "loading %s profile from %s" % (title, directory)
        self.directory = directory
        self._title = title
        if not os.path.exists(directory):
            os.makedirs(self.directory)
        self._create_missing_content()
        self._bookmarks = None
        self._tools = None

    def title(self):
        return self._title

    def _create_missing_content(self):
        required = [
            BOOKMARKS_FILENAME,
            TOOLS_FILENAME,
            ]
        for filename in required:
            self._create_file_if_missing(filename)

    def _create_file_if_missing(self, name):
        pathname = osutils.pathjoin(self.directory, name)
        if not os.path.exists(pathname):
            # TODO: copy content from a template directory
            content = _DEFAULT_CONTENT.get(name, "").lstrip()
            f = open(pathname, "w")
            f.write(content)
            f.close()

    def bookmarks(self):
        if self._bookmarks is None:
            self._bookmarks = bookmarks.BookmarkStore(self.bookmarks_path())
        return self._bookmarks

    def tools(self):
        if self._tools is None:
            self._tools = tools.ToolStore(self.tools_path())
        return self._tools

    def custom_editors(self, desktop_name):
        editors = {}
        cfg_path = self.editors_path()
        cfg = config.ConfigObj(cfg_path, encoding="utf-8")
        for section in ['DEFAULT', desktop_name]:
            try:
                data = cfg[section]
            except KeyError:
                continue
            else:
                for key, value in data.items():
                    exts = key.split(',')
                    for ext in exts:
                        editors[ext.strip()] = value
        return editors

    def bookmarks_path(self):
        return osutils.pathjoin(self.directory, BOOKMARKS_FILENAME)

    def tools_path(self):
        return osutils.pathjoin(self.directory, TOOLS_FILENAME)

    def editors_path(self):
        return osutils.pathjoin(self.directory, EDITORS_FILENAME)


def load_profile(name=None):
    """Return the Explorer profile with a given name.
    
    :param name: if None, a default profile is returned.
    """
    profile_dir = config_dir()
    if name is None:
        title = gettext("My")
    else:
        profile_basename = "%s-hat" % (name,)
        profile_dir = osutils.pathjoin(profile_dir, profile_basename)
        title = name
    return ExplorerProfile(profile_dir, title)


def find_available_profile_names():
    """Find the names of available profiles."""
    base_dir = config_dir()
    result = []
    for name in os.listdir(base_dir):
        if name.endswith("-hat"):
            fullpath = osutils.pathjoin(base_dir, name)
            if osutils.isdir(fullpath):
                result.append(name[:-4])
    return sorted(result)


def config_dir():
    """Get the configuration directory for Explorer."""
    return osutils.pathjoin(config.config_dir(), "explorer")
