# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys

from PyQt4 import QtCore, QtGui

try:
    from bzrlib.osutils import parent_directories
except ImportError:
    from bzrlib.plugins.explorer.lib.helpers import parent_directories

from bzrlib.plugins.explorer.lib.helpers import build_combo_with_labels
from bzrlib.plugins.explorer.lib.i18n import gettext, N_
from bzrlib.plugins.explorer.lib.kinds import (
    CANCEL_ACTION,
    OPEN_FOLDER_ACTION,
    EDIT_ACTION,
    icon_for_kind,
    )
from bzrlib.plugins.qbzr.lib import treewidget


class WorkingTreeBrowser(QtGui.QDockWidget):

    def __init__(self, action_callback, browse_action, *args):
        self._action_callback = action_callback
        self._browse_action = browse_action
        QtGui.QDockWidget.__init__(self, *args)
        self.setObjectName("WorkingTreeDockWidget")
        self._style = 'qbrowse'
        self._tree = None
        self._branch = None

        # Show either a message or the actual browser
        msg = QtGui.QTextBrowser()
        msg.setText(gettext("No working tree"))
        self._browser = _QBrowseStyleBrowser(action_callback, browse_action, self)
        self._stacked = QtGui.QStackedWidget()
        self._stacked.addWidget(msg)
        self._stacked.addWidget(self._browser)
        self.setWidget(self._stacked)

    def set_tree(self, tree, branch):
        """Display the given working tree.

        :param tree: the WorkingTree to display.
            If None, the view is cleared.
        """
        if tree is None:
            # Show the message
            self._stacked.setCurrentIndex(0)
        else:
            # Show the browser
            self._browser.set_tree(tree, branch)
            if self._style == 'qbrowse':
                index = 1
            else:
                index = 2
            self._stacked.setCurrentIndex(index)
        self._tree = tree
        self._branch = branch

    def refresh(self):
        """Refresh the browser."""
        self._browser.refresh()

    def get_style(self):
        return self._style

    def set_style(self, style):
        """Set the style: either qbrowse or classic."""
        self._style = style
        if style == 'classic':
            if self._stacked.count() == 2:
                classic_browser = _ClassicBrowser(self._action_callback,
                    self._browse_action)
                self._stacked.addWidget(classic_browser)
            index = 2
        else:
            index = 1
        self._browser = self._stacked.widget(index)
        self._browser.set_tree(self._tree, self._branch)
        self._stacked.setCurrentIndex(index)


class _ActionPanel(QtGui.QDialogButtonBox):

    def __init__(self, action_callback, browse_action, *args):
        self._action_callback = action_callback
        self._browse_action = browse_action
        QtGui.QDialogButtonBox.__init__(self, *args)
        self._root = None
        self._selected_fileinfo = None

        # Build the buttons and add them to a panel
        browse_button = QtGui.QPushButton()
        browse_button.setText(self._browse_action.iconText())
        browse_button.setIcon(self._browse_action.icon())
        open_button = QtGui.QPushButton()
        open_button.setText(gettext("&Manage"))
        open_button.setIcon(icon_for_kind(OPEN_FOLDER_ACTION))
        edit_button = QtGui.QPushButton()
        edit_button.setText(gettext("&Edit"))
        edit_button.setIcon(icon_for_kind(EDIT_ACTION))
        self.addButton(browse_button, QtGui.QDialogButtonBox.ActionRole)
        self.addButton(open_button, QtGui.QDialogButtonBox.ActionRole)
        self.addButton(edit_button, QtGui.QDialogButtonBox.AcceptRole)
        self.connect(browse_button, QtCore.SIGNAL("clicked(bool)"),
            self._do_browse_items)
        self.connect(open_button, QtCore.SIGNAL("clicked(bool)"),
            self._do_open_folder)
        self.connect(edit_button, QtCore.SIGNAL("clicked(bool)"),
            self._do_edit_file)
        self._edit_button = edit_button

    def _do_browse_items(self):
        self._action_callback("browse", None)

    def _do_open_folder(self):
        fileinfo = self._selected_fileinfo
        if fileinfo is None:
            folder = self._root
        elif fileinfo.isDir():
            folder = fileinfo.canonicalFilePath()
        else:
            folder = fileinfo.canonicalPath()
        self._action_callback("open", folder)

    def _do_edit_file(self):
        fileinfo = self._selected_fileinfo
        path = fileinfo.canonicalFilePath()
        self._action_callback("edit", unicode(path))

    def set_root(self, path):
        self._root = path

    def set_selection(self, fileinfo):
        self._selected_fileinfo = fileinfo
        edit_is_enabled = True
        if fileinfo is None or fileinfo.isDir():
            edit_is_enabled = False
        self._edit_button.setEnabled(edit_is_enabled)


class _FilterPanel(QtGui.QWidget):

    def __init__(self, *args):
        QtGui.QWidget.__init__(self, *args)
        self._tree_viewer = None
        self._current_filter = ''

        # Build the controls
        filter_label = QtGui.QLabel(gettext("Filter:"))
        self._filter_field = QtGui.QLineEdit()
        self._filter_cancel_button = QtGui.QToolButton()
        self._filter_cancel_button.setAutoRaise(True)
        self._filter_cancel_button.setEnabled(False)
        self._filter_cancel_button.setIcon(icon_for_kind(CANCEL_ACTION))
        self._filter_cancel_button.setToolTip(gettext("Clear the filter"))

        # Connect up the controls to handlers
        self.connect(self._filter_field, QtCore.SIGNAL("textChanged(QString)"),
            self._apply_filter)
        self.connect(self._filter_cancel_button, QtCore.SIGNAL("clicked(bool)"),
            self.clear_filter)

        # Put the controls together
        layout = QtGui.QHBoxLayout()
        layout.addWidget(filter_label)
        layout.addWidget(self._filter_field)
        layout.addWidget(self._filter_cancel_button)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def _apply_filter(self, text):
        self._tree_viewer.invalidate_filter()
        if text:
            self._filter_cancel_button.setEnabled(True)
            # Filtering active. We want to see all matches and
            # this mostly works. It seems to struggle though when
            # the filter gets shorter - maybe a Qt 4.4 bug w.r.t.
            # propagating exactly what's changed to the view?
            self._tree_viewer.expandAll()
        else:
            # Filtering cleared. Restore the initial view with
            # everything collapsed. (If the user has selectively
            # expanded some directories out before filtering,
            # collapsing all here implies losing that setup. We
            # could always remember the expanded directories at
            # the start of filtering and restore them at this
            # point if it proves a problem in practice.)
            self._tree_viewer.collapseAll()
            self._tree_viewer.refresh()
            self._filter_cancel_button.setEnabled(False)
        self._current_filter = text

    def clear_filter(self):
        self._filter_field.setText('')

    def get_filter(self):
        return unicode(self._filter_field.text())

    def set_tree_viewer(self, tree_viewer):
        self._tree_viewer = tree_viewer


class _TreeViewer(QtGui.QTreeView):

    def __init__(self, action_callback, selection_callback,
            filter_callback, *args):
        self._action_callback = action_callback
        self._selection_callback = selection_callback
        self._filter_callback = filter_callback
        QtGui.QTreeView.__init__(self, *args)

        # Use a proxy model so we can filter on the path.
        self._source_model = _TreeModel()
        self._model = _PathBasedSortFilterProxyModel(self._filter_callback)
        self._model.setSourceModel(self._source_model)
        self.setModel(self._model)
        self._configure_ui()

        # Track the selection
        self.connect(self, QtCore.SIGNAL("doubleClicked(QModelIndex)"),
            self._do_double_clicked)
        sel_model = self.selectionModel()
        self.connect(sel_model,
            QtCore.SIGNAL("currentChanged(QModelIndex,QModelIndex)"),
            self._do_current_changed)

    def _configure_ui(self):
        # Hide all except the Name column
        for i in  range(1, self._model.columnCount()):
            self.setColumnHidden(i, True)
        if getattr(self, 'setHeaderHidden', None):
            self.setHeaderHidden(True)

    def _do_double_clicked(self, model_index):
        fileinfo = self._get_fileinfo(model_index)
        if fileinfo is not None and not fileinfo.isDir():
            path = fileinfo.canonicalFilePath()
            self._action_callback("open", unicode(path))

    def _do_current_changed(self, model_index, prev_model_index):
        fileinfo = self._get_fileinfo(model_index)
        self._selection_callback(fileinfo)

    def _get_fileinfo(self, model_index):
        model = model_index.model()
        if model is None:
            fileinfo = None
        else:
            source_index = model.mapToSource(model_index)
            fileinfo = self._source_model.fileInfo(source_index)
        return fileinfo

    def set_tree(self, tree, branch):
        """Change the view to display the given working tree."""
        path = tree.basedir
        source_index = self._source_model.index(path)
        index = self._model.mapFromSource(source_index)
        self.setRootIndex(index)
        self.selectionModel().reset()
        self._model.set_root(path)

    def refresh(self):
        """Refresh the view."""
        self._source_model.refresh()

    def invalidate_filter(self):
        self._model.invalidateFilter()


class _TreeModel(QtGui.QDirModel):

    def __init__(self, parent=None):
        QtGui.QDirModel.__init__(self, parent)
        # Set the sort flags: directories first,
        # on Windows items are case-insensitive
        flags = QtCore.QDir.SortFlags(QtCore.QDir.Name|QtCore.QDir.DirsFirst)
        if sys.platform == 'win32':
            flags |= QtCore.QDir.IgnoreCase
        self.setSorting(flags)


class _PathBasedSortFilterProxyModel(QtGui.QSortFilterProxyModel):

    def __init__(self, filter_callback, parent=None):
        QtGui.QSortFilterProxyModel.__init__(self, parent)
        self._filter_callback = filter_callback
        self._root = None
        self._interesting = None

    def set_root(self, root):
        self._root = root

    def invalidateFilter(self):
        self._text = self._filter_callback()
        if self._text == '':
            self._interesting = None
        else:
            # We only want to show directories containing matching items so
            # we need to do a pass in advance working out what's of interest
            self._source = self.sourceModel()
            root_index = self._source.index(self._root)
            self._interesting = self._find_interesting(self._source,
                root_index, self._text)
        QtGui.QSortFilterProxyModel.invalidateFilter(self)

    def _find_interesting(self, source, root_index, text):
        interesting = set([''])
        directories = set()
        for source_index in self._iter_children(source, root_index):
            fileinfo = source.fileInfo(source_index)
            abspath = unicode(fileinfo.absoluteFilePath())
            relpath = abspath[len(self._root) + 1:]
            if relpath.find(text) >= 0:
                interesting.add(relpath)
                directories.update(parent_directories(relpath))
        interesting.update(directories)
        #print "\ninteresting for %s ...\n%s" % (text, "\n".join(sorted(interesting)))
        return interesting

    def _iter_children(self, source, parent):
        for i in range(0, source.rowCount(parent)):
            index = source.index(i, 0, parent)
            yield index
            for child in self._iter_children(source, index):
                yield child
        
    def filterAcceptsRow(self, source_row, source_parent):
        if self._interesting is None:
            return True
        result = False
        source_index = self._source.index(source_row, 0, source_parent)
        fileinfo = self._source.fileInfo(source_index)
        abspath = unicode(fileinfo.absoluteFilePath())
        if abspath.startswith(self._root):
            # We're inside the tree of interest
            relpath = abspath[len(self._root) + 1:]
            result = relpath in self._interesting
        elif fileinfo.isDir():
            # We're interested if it's a parent of the root
            result = self._root.startswith(abspath)
        return result


class _ClassicBrowser(QtGui.QWidget):

    filter_panel_factory = _FilterPanel
    tree_viewer_factory = _TreeViewer
    action_panel_factory = _ActionPanel

    def __init__(self, action_callback, browse_action, *args):
        self._action_callback = action_callback
        self._browse_action = browse_action
        QtGui.QWidget.__init__(self, *args)

        # Build the ui components
        self._action_panel = self.action_panel_factory(action_callback,
            browse_action)
        self._filter_panel = self.filter_panel_factory()
        self._tree_viewer = self.tree_viewer_factory(action_callback,
            self._action_panel.set_selection, self._filter_panel.get_filter)
        self._filter_panel.set_tree_viewer(self._tree_viewer)

        # Put them together
        layout = QtGui.QVBoxLayout()
        layout.addWidget(self._filter_panel)
        layout.addWidget(self._tree_viewer)
        layout.addWidget(self._action_panel)
        self.setLayout(layout)

    def set_tree(self, tree, branch):
        """Display the given working tree."""
        self._tree_viewer.set_tree(tree, branch)
        self._action_panel.set_root(tree.basedir)
        self._action_panel.set_selection(None)
        self._filter_panel.clear_filter()

    def refresh(self):
        """Refresh the browser."""
        self._tree_viewer.refresh()


### QBrowse style browser support ###

# The map from filter category to filter flags for TreeFilterProxyModels
_FILTER_FLAGS_MAP = {
    # The flag order is UNCHANGED, CHANGED, UNVERSIONED, IGNORED
    'versioned':        [True, True, True, False],
    'changed':          [False, True, False, False],
    'ignored':          [False, False, True, True],
    'unversioned':      [False, False, True, False],
    'all':              [True, True, True, True],
    }

class _QBrowseStyleFilterPanel(QtGui.QWidget):

    def __init__(self, *args):
        QtGui.QWidget.__init__(self, *args)
        self._tree_viewer = None

        # Build the controls
        self._category_combo = build_combo_with_labels([
            ('versioned', gettext("Versioned")),
            ('changed', gettext("Changed")),
            ('ignored', gettext("Ignored")),
            ('unversioned', gettext("Unversioned")),
            ('all', gettext("All")),
            ])

        # Connect up the controls to handlers
        QtCore.QObject.connect(self._category_combo,
            QtCore.SIGNAL("currentIndexChanged(int)"),
            self._apply_category)

        # Put the controls together
        layout = QtGui.QHBoxLayout()
        layout.addWidget(self._category_combo)
        layout.addStretch(1)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def _apply_category(self, index):
        category = str(self._category_combo.itemData(index).toString())
        filter_flags = _FILTER_FLAGS_MAP[category]
        #print "category %s mapped to flags %s" % (category, filter_flags)
        self._tree_viewer.tree_filter_model.setFilters(filter_flags)

    def clear_filter(self):
        # do nothing?
        pass

    def get_filter(self):
        index = self._category_combo.currentIndex()
        return str(self._category_combo.itemData(index).toString())

    def set_tree_viewer(self, tree_viewer):
        self._tree_viewer = tree_viewer


class _QBrowseStyleTreeViewer(treewidget.TreeWidget):

    def __init__(self, action_callback, selection_callback,
            filter_callback, *args):
        self._selection_callback = selection_callback
        treewidget.TreeWidget.__init__(self, *args)

        # Track the selection
        self.connect(self.selectionModel(),
            QtCore.SIGNAL("currentChanged(QModelIndex,QModelIndex)"),
            self._do_current_changed)

    def _do_current_changed(self, model_index, prev_model_index):
        item_data = self.get_selection_items([model_index])[0]
        fileinfo = QtCore.QFileInfo(item_data.path)
        self._selection_callback(fileinfo)

    def invalidate_filter(self):
        self.tree_filter_model.invalidateFilter()


class _QBrowseStyleBrowser(_ClassicBrowser):

    filter_panel_factory = _QBrowseStyleFilterPanel
    tree_viewer_factory = _QBrowseStyleTreeViewer
